import { SQL, count, desc, eq, inArray, or, sql } from 'drizzle-orm';
import { DB, auditLog, user, type AuditLog, type NewAuditLog, type User } from '../drizzle';
import {
	AuditAction,
	type AuditListItem,
	type AuditSearchClause,
	type MinimalRequestEvent
} from './types';
import type { Paginated, PaginationMeta } from '$lib/types';

const AUTOFLAG = [
	AuditAction.MALICIOUS_REQUEST,
	AuditAction.THROTTLE,
	AuditAction.DEACTIVATION_REQUEST,
	AuditAction.DATA_DOWNLOAD_REQUEST
];

export class Audit {
	public static async insert(
		action: AuditAction,
		comment?: string,
		user?: User,
		ip?: string,
		userAgent?: string
	) {
		const newAuditLog: NewAuditLog = {
			action,
			content: comment,
			actorId: user?.id,
			actor_ip: ip,
			actor_ua: userAgent,
			flagged: Number(AUTOFLAG.includes(action))
		};

		// TODO: send flagged to administrator

		await DB.drizzle.insert(auditLog).values(newAuditLog);
	}

	public static async insertRequest(
		action: AuditAction,
		{ request, getClientAddress }: MinimalRequestEvent,
		user?: User,
		comment?: string
	) {
		const ip = getClientAddress();
		const userAgent = request.headers.get('user-agent') || undefined;

		return Audit.insert(action, comment, user, ip, userAgent);
	}

	public static async auditCount(search: AuditSearchClause) {
		const [{ rowCount }] = await DB.drizzle
			.select({ rowCount: count(auditLog.id).mapWith(Number) })
			.from(auditLog)
			.leftJoin(user, eq(user.id, auditLog.actorId))
			.where(Audit.getAuditWhere(search));
		return rowCount;
	}

	public static async searchAudit(
		search: AuditSearchClause & {
			limit?: number;
			offset?: number;
		}
	) {
		const limit = search.limit || 20;
		const rowCount = await Audit.auditCount(search);

		const auditSubquery = DB.drizzle
			.select({ id: auditLog.id })
			.from(auditLog)
			.leftJoin(user, eq(user.id, auditLog.actorId))
			.limit(limit)
			.offset(search.offset || 0)
			.where(Audit.getAuditWhere(search))
			.orderBy(desc(auditLog.created_at))
			.as('auditSubquery');

		const junkList = await DB.drizzle
			.select({
				audit_log: auditLog,
				user: user
			})
			.from(auditSubquery)
			.innerJoin(auditLog, eq(auditLog.id, auditSubquery.id))
			.leftJoin(user, eq(user.id, auditLog.actorId));

		const list = Audit.mapAuditRows(junkList);
		const meta: PaginationMeta = {
			rowCount,
			pageSize: limit,
			pageCount: Math.ceil(rowCount / limit)
		};

		return { list, meta } satisfies Paginated<AuditListItem>;
	}

	private static getAuditWhere(search: AuditSearchClause) {
		const selectList: SQL<unknown>[] = [];
		if (search.actions?.length) {
			const actions = Array.isArray(search.actions)
				? search.actions
				: (search.actions as string).split(',');
			selectList.push(inArray(auditLog.action, actions));
		}

		if (search.content) {
			selectList.push(sql`lower(${auditLog.content}) LIKE ${`%${search.content.toLowerCase()}%`}`);
		}

		if (search.ip) {
			selectList.push(sql`lower(${auditLog.actor_ip}) LIKE ${`%${search.ip.toLowerCase()}%`}`);
		}

		if (search.ua) {
			selectList.push(sql`lower(${auditLog.actor_ua}) LIKE ${`%${search.ua.toLowerCase()}%`}`);
		}

		if (search.user) {
			selectList.push(eq(user.uuid, search.user));
		}

		if (search.flagged) {
			selectList.push(eq(auditLog.flagged, 1));
		}

		return or(...selectList);
	}

	private static mapAuditRows(
		rows: {
			audit_log: AuditLog;
			user?: User | null;
		}[]
	) {
		return rows.reduce<AuditListItem[]>((accum, entry) => {
			let existingEntry = accum.find((e) => e.id === entry.audit_log.id);
			if (!existingEntry) {
				existingEntry = {
					id: entry.audit_log.id,
					action: entry.audit_log.action as AuditAction,
					ip: entry.audit_log.actor_ip || undefined,
					ua: entry.audit_log.actor_ua || undefined,
					content: entry.audit_log.content || undefined,
					flagged: Boolean(entry.audit_log.flagged),
					created_at: entry.audit_log.created_at
				};
				accum.push(existingEntry);
			}

			if (entry.user) {
				existingEntry.user = {
					uuid: entry.user.uuid,
					name: entry.user.display_name
				};
			}

			return accum;
		}, []);
	}
}
